﻿' (C) Copyright 2010 by Autodesk, Inc. 
'
Imports System
Imports Autodesk.AutoCAD.Runtime
Imports Autodesk.AutoCAD.ApplicationServices
Imports Autodesk.AutoCAD.DatabaseServices
Imports Autodesk.AutoCAD.Geometry
Imports Autodesk.AutoCAD.EditorInput

' This line is not mandatory, but improves loading performances
<Assembly: CommandClass(GetType(HowToDisposeSample.MyCommands))> 

Namespace HowToDisposeSample

 
    Public Class MyCommands

        ' Not a good way
        <CommandMethod("MyGroup", "Test1", "MyCommandLocal", CommandFlags.Modal)> _
        Public Sub MyCommand()
            Dim trans As Transaction
            Dim myLine As Line
            Try
                Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
                trans = db.TransactionManager.StartTransaction
                Dim btr As BlockTableRecord = trans.GetObject(db.CurrentSpaceId, OpenMode.ForWrite)
                myLine = New Line(New Point3d(0, 0, 0), New Point3d(100, 100, 0))
                Throw New Autodesk.AutoCAD.Runtime.Exception(ErrorStatus.AtMaxReaders)
                btr.AppendEntity(myLine)
                trans.AddNewlyCreatedDBObject(myLine, True)
                trans.Commit()
            Catch ex As Autodesk.AutoCAD.Runtime.Exception
                'Do something to handle the error
            Finally
                If Not trans Is Nothing Then
                    trans.Dispose()
                End If
                'If Not myLine Is Nothing Then
                '    myLine.Dispose()
                'End If
            End Try
        End Sub

        'A common mistake
        <CommandMethod("MyGroup", "Test2", "MyCommandLocal2", CommandFlags.Modal)> _
        Public Sub MyCommand2()
            Try
                Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
                Using trans As Transaction = db.TransactionManager.StartTransaction
                    Dim btr As BlockTableRecord = trans.GetObject(db.CurrentSpaceId, OpenMode.ForWrite)
                    Dim myLine As New Line(New Point3d(0, 0, 0), New Point3d(100, 100, 0))
                    Throw New Autodesk.AutoCAD.Runtime.Exception(ErrorStatus.AtMaxReaders)
                    btr.AppendEntity(myLine)
                    trans.AddNewlyCreatedDBObject(myLine, True)
                    trans.Commit()
                End Using
            Catch ex As Autodesk.AutoCAD.Runtime.Exception
                'Do something to handle the error
            Finally
                'Cleanup here - nothing is needed in this case
            End Try
        End Sub

        'The best way
        <CommandMethod("MyGroup", "Test3", "MyCommandLocal3", CommandFlags.Modal)> _
        Public Sub MyCommand3()
            Try
                Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
                Using trans As Transaction = db.TransactionManager.StartTransaction
                    Dim btr As BlockTableRecord = trans.GetObject(db.CurrentSpaceId, OpenMode.ForWrite)
                    Using myLine As New Line(New Point3d(0, 0, 0), New Point3d(100, 100, 0))
                        btr.AppendEntity(myLine)
                        trans.AddNewlyCreatedDBObject(myLine, True)
                    End Using
                    trans.Commit()
                End Using
            Catch ex As Autodesk.AutoCAD.Runtime.Exception
                'Do something to handle the error
            Finally
                'Cleanup here - nothing is needed in this case
            End Try
        End Sub

        'The best way - simulating what happens when there's an error
        <CommandMethod("MyGroup", "Test4", "MyCommandLocal4", CommandFlags.Modal)> _
        Public Sub MyCommand4()
            Try
                Dim db As Database = Application.DocumentManager.MdiActiveDocument.Database
                Using trans As Transaction = db.TransactionManager.StartTransaction
                    Dim btr As BlockTableRecord = trans.GetObject(db.CurrentSpaceId, OpenMode.ForWrite)
                    Using myLine As New Line(New Point3d(0, 0, 0), New Point3d(100, 100, 0))
                        btr.AppendEntity(myLine)
                        Throw New Autodesk.AutoCAD.Runtime.Exception(ErrorStatus.AtMaxReaders)
                        trans.AddNewlyCreatedDBObject(myLine, True)
                    End Using
                    trans.Commit()
                End Using
            Catch ex As Autodesk.AutoCAD.Runtime.Exception
                'Do something to handle the error
            Finally
                'Cleanup here - nothing is needed in this case
            End Try
        End Sub



        'Overkill - but still ok.
        <CommandMethod("MyGroup", "Test5", "MyCommandLocal5", CommandFlags.Modal)> _
        Public Sub MyCommand5()
            Using db As Database = Application.DocumentManager.MdiActiveDocument.Database
                Using trans As Transaction = db.TransactionManager.StartTransaction
                    Using btr As BlockTableRecord = trans.GetObject(db.CurrentSpaceId, OpenMode.ForWrite)
                        Using myLine As New Line(New Point3d(0, 0, 0), New Point3d(100, 100, 0))
                            btr.AppendEntity(myLine)
                            trans.AddNewlyCreatedDBObject(myLine, True)
                        End Using
                        For Each id As ObjectId In btr
                            Using ent As Entity = trans.GetObject(id, OpenMode.ForRead)
                                Application.DocumentManager.MdiActiveDocument.Editor.WriteMessage(vbLf & ent.ObjectId.ToString)
                            End Using
                        Next
                        trans.Commit()
                    End Using
                End Using
            End Using
        End Sub
    End Class

End Namespace