// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// Cyrille Fauvel
// Autodesk Developer Network

//---------------------------------------------------------------------
//- asdkOPLine.cpp : Implementation of asdkOPLine
//---------------------------------------------------------------------
#include "StdAfx.h"
#include "asdkOPLine.h"
#include "asdkOPLineStyle.h"

//---------------------------------------------------------------------
Adesk::UInt32 asdkOPLine::kCurrentVersionNumber =1 ;

//---------------------------------------------------------------------
#define ARXOK(a) {\
	Acad::ErrorStatus es ;\
	if ( (es =a) != Acad::eOk )\
		return (es) ;\
	}
#define ARXOK2(a) {\
	Acad::ErrorStatus es ;\
	if ( (es =a) != Acad::eOk && es != Acad::eObjectToBeDeleted )\
		return (es) ;\
	}

//---------------------------------------------------------------------
ACRX_DXF_DEFINE_MEMBERS (
	asdkOPLine, AcDbPolyline,
	AcDb::kDHL_CURRENT, AcDb::kMReleaseCurrent, 
	AcDbProxyEntity::kNoOperation, ASDKOPLINE,
ASDKOPLINEAPP
|Product Desc:     A description for your object
|Company:          Your company name
|WEB Address:      Your company WEB site address
)

//---------------------------------------------------------------------
//asdkOPLine::asdkOPLine () : AcDbPolyline () {}

asdkOPLine::asdkOPLine (AcDbObjectId id)
	: AcDbPolyline (), m_OPLineStyleId(id)
{
}

asdkOPLine::asdkOPLine (unsigned int num_verts, AcDbObjectId id)
	: AcDbPolyline (num_verts), m_OPLineStyleId(id)
{
}

asdkOPLine::~asdkOPLine () {
}

//---------------------------------------------------------------------
Acad::ErrorStatus asdkOPLine::oPLineStyle (AcDbHardPointerId &OPLineStyle) {
	assertReadEnabled () ;
	OPLineStyle =m_OPLineStyleId ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLine::setOPLineStyle (const AcDbHardPointerId OPLineStyle) {
	assertWriteEnabled () ;
	m_OPLineStyleId =OPLineStyle ;
	return (Acad::eOk) ;
}

//---------------------------------------------------------------------
Acad::ErrorStatus asdkOPLine::makeFrom (AcDbObjectId id) {
	assertWriteEnabled () ;
	AcDbPolyline *pLine =NULL ;
	//- Since we handover to the entity, we do not use smartpointers here
	ARXOK ( acdbOpenObject (pLine, id, AcDb::kForRead) ) ;
	ARXOK2 ( makeFrom (pLine) ) ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLine::makeFrom (AcDbPolyline *pLine) {
	assertWriteEnabled () ;
	assert ( pLine != NULL ) ;
	if ( pLine == NULL )
		return (Acad::eInvalidInput) ;
	//- Entity
	setLayer (pLine->layerId ()) ;
	setLinetype (pLine->linetypeId ()) ;
	setLinetypeScale (pLine->linetypeScale ()) ;
	setLineWeight (pLine->lineWeight ()) ;
	setColorIndex (pLine->colorIndex ()) ;
	setVisibility (pLine->visibility ()) ;
	//- Pline
	double d, w0, w1 ;
	AcGePoint2d pt ;
	reset (false, 0) ;
	for ( unsigned int i =0 ; i < pLine->numVerts () ; i++ ) {
		pLine->getBulgeAt (i, d) ;
		pLine->getWidthsAt (i, w0, w1) ;
		pLine->getPointAt (i, pt) ;
		addVertexAt (i, pt, d, w0, w1) ;
	}
	setClosed (pLine->isClosed ()) ;
	if ( pLine->getConstantWidth (d) == Acad::eOk )
		setConstantWidth (d) ;
	setElevation (pLine->elevation ()) ;
	setNormal (pLine->normal ()) ;
	setPlinegen (pLine->hasPlinegen ()) ;
	setThickness (pLine->thickness ()) ;

	if ( !pLine->isWriteEnabled () )
		ARXOK ( pLine->upgradeOpen () ) ;
	ARXOK2 ( pLine->handOverTo (this, true, true) ) ;
	delete pLine ;
	return (Acad::eOk) ;
}

//---------------------------------------------------------------------
//- AcDbObject protocols
//- Dwg Filing protocol
Acad::ErrorStatus asdkOPLine::dwgOutFields (AcDbDwgFiler *pFiler) const {
	assertReadEnabled () ;
	//- Save parent class information first.
	Acad::ErrorStatus es =AcDbPolyline::dwgOutFields (pFiler) ;
	if ( es != Acad::eOk )
		return (es) ;
	//- Object version number needs to be saved first
	if ( (es =pFiler->writeUInt32 (asdkOPLine::kCurrentVersionNumber)) != Acad::eOk )
		return (es) ;
	//- Output params
	//.....
	pFiler->writeItem (m_OPLineStyleId) ;

	return (pFiler->filerStatus ()) ;
}

Acad::ErrorStatus asdkOPLine::dwgInFields (AcDbDwgFiler *pFiler) {
	assertWriteEnabled () ;
	//- Read parent class information first.
	Acad::ErrorStatus es =AcDbPolyline::dwgInFields (pFiler) ;
	if ( es != Acad::eOk )
		return (es) ;
	//- Object version number needs to be read first
	Adesk::UInt32 version =0 ;
	if ( (es =pFiler->readUInt32 (&version)) != Acad::eOk )
		return (es) ;
	if ( version > asdkOPLine::kCurrentVersionNumber )
		return (Acad::eMakeMeProxy) ;
	//- Uncomment the 2 following lines if your current object implementation cannot
	//- support previous version of that object.
	//if ( version < asdkOPLine::kCurrentVersionNumber )
	//	return (Acad::eMakeMeProxy) ;
	//- Read params
	//.....
	pFiler->readItem (&m_OPLineStyleId) ;

	return (pFiler->filerStatus ()) ;
}

Acad::ErrorStatus asdkOPLine::subDeepClone (
	AcDbObject *pOwnerObject,
	AcDbObject *&pClonedObject,
	AcDbIdMapping &idMap,
	Adesk::Boolean isPrimary
) const {
	assertReadEnabled () ;
	Acad::ErrorStatus es =AcDbObject::subDeepClone (pOwnerObject, pClonedObject, idMap, isPrimary) ;

	if (   idMap.deepCloneContext () == AcDb::kDcInsert
		|| idMap.deepCloneContext () == AcDb::kDcInsertCopy
	) {
		//- We should force the copy of the style
		if ( !idMap.compute (AcDbIdPair (m_OPLineStyleId, AcDbObjectId::kNull, true)) ) {
			//- Open the orignal dictionary object we want to clone
			AcDbObjectPointer<asdkOPLineStyle> pObj (m_OPLineStyleId, AcDb::kForRead) ;
			//- Open the dictionary owner of the original dictionary object
			AcDbDictionaryPointer pDict (pObj->ownerId (), AcDb::kForRead) ;
			//- Clone the dictionary owner if necessary, if not open it for write
			AcDbDatabase *pDestDb =NULL ;
			idMap.destDb (pDestDb) ;
			AcDbDictionaryPointer pNOD (pDestDb->namedObjectsDictionaryId (), AcDb::kForRead) ;
			if ( pNOD->has (ASDK_OPLINESTYLE) ) {
				//- It is there
				AcDbObjectId idc ;
				pNOD->getAt (ASDK_OPLINESTYLE, idc) ;
				AcDbDictionaryPointer pClonedDict (idc, AcDb::kForWrite) ;
				//- Clone the dictionary object now
				asdkOPLineStyle *pClone =NULL ;
				pObj->deepClone (pClonedDict, (AcDbObject *&)pClone, idMap, isPrimary) ;
				//- Assign its owner ourself with an anonymous key
				pClonedDict->setAt (_T("*"), pClone, idc) ;
				//- We *must* close the clones ourself
				pClone->close () ;
			} else {
				//- Clone it
				pNOD->upgradeOpen () ;
				AcDbDictionary *pClonedDict =NULL ;
				pDict->deepClone (pNOD, (AcDbObject *&)pClonedDict, idMap, isPrimary) ;
				AcDbObjectId idc ;
				pNOD->setAt (ASDK_OPLINESTYLE, pClonedDict, idc) ;
				pNOD->downgradeOpen () ;
				//- We *must* close the clones ourself
				pClonedDict->close () ;
			}
		}
	}
	return (es) ;
}

//---------------------------------------------------------------------
//- AcDbEntity protocols
Adesk::Boolean asdkOPLine::subWorldDraw (AcGiWorldDraw *mode) {
	assertReadEnabled () ;
	AcDbPolyline::subWorldDraw (mode) ;

	TCHAR *pltName =plotStyleName () ;
	AcDbDictionaryPointer pNOD (acdbHostApplicationServices ()->workingDatabase ()->namedObjectsDictionaryId (), AcDb::kForRead) ;
	AcDbObjectId pltStyleNameId ;
	pNOD->getAt (_T("ACAD_PLOTSTYLENAME"), pltStyleNameId) ;
	AcDbDictionaryPointer pPltStyleName (pltStyleNameId, AcDb::kForRead) ;
	pPltStyleName->getAt (pltName, pltStyleNameId) ;
	mode->subEntityTraits ().setPlotStyleName (AcDb::kPlotStyleNameById, pltStyleNameId) ;

	AcDbObjectPointer<asdkOPLineStyle> pStyle (m_OPLineStyleId, AcDb::kForRead) ;
	if ( pStyle.openStatus () == Acad::eOk ) {
		for ( Adesk::UInt32 i =0 ; i < pStyle->numOffsets () ; i++ ) {
			//- Create offset
			AcDbVoidPtrArray vpArr ;
			double off ;
			pStyle->offsetOf (i, off) ;
			if ( off == 0 )
				continue ;
			if ( AcDbPolyline::getOffsetCurves (off, vpArr) != Acad::eOk )
				continue ;

			for ( int j =0 ; j < vpArr.length () ; j++ ) {
				AcDbEntity *pCopy =(AcDbEntity *)vpArr [j] ;
				//- Entity
				pCopy->setLayer (layerId ()) ;
				pCopy->setLinetypeScale (linetypeScale ()) ;
				pCopy->setVisibility (visibility ()) ;
				//- Set properties based on style
				Adesk::UInt16 color ;
				AcDbObjectId ltId ;
				AcDb::LineWeight lwt ;
				pStyle->colorOf (i, color) ;
				pStyle->linetypeOf (i, ltId) ;
				pStyle->lineWeightOf (i, lwt) ;
				pCopy->setColorIndex (color) ;
				pCopy->setLinetype (ltId) ;
				pCopy->setLineWeight (lwt) ;
				//- Polyline
				if ( pCopy->isKindOf (AcDbPolyline::desc ()) )
					((AcDbPolyline *)pCopy)->setElevation (elevation ()) ;

				pCopy->setPlotStyleName (pltName) ;

				mode->subEntityTraits ().setColor (color) ;
				mode->subEntityTraits ().setThickness (thickness ()) ;
				mode->subEntityTraits ().setLayer (layerId ()) ;
				mode->subEntityTraits ().setLineType (ltId) ;
				mode->subEntityTraits ().setLineTypeScale (linetypeScale ()) ;
				mode->subEntityTraits ().setLineWeight (lwt) ;
				mode->subEntityTraits ().setPlotStyleName (AcDb::kPlotStyleNameById, pltStyleNameId) ;

				pCopy->worldDraw (mode) ;
				delete pCopy ;
			}
		}
	}
	delete [] pltName ;

	return (Adesk::kTrue) ;
}

Adesk::UInt32 asdkOPLine::subSetAttributes (AcGiDrawableTraits *traits) {
	assertReadEnabled () ;
	return (AcDbPolyline::subSetAttributes (traits)) ;
}

Acad::ErrorStatus asdkOPLine::subExplode (AcDbVoidPtrArray &entitySet) const {
	assertReadEnabled () ;
	AcDbPolyline *pCopy =new AcDbPolyline ;
	//- Entity
	pCopy->setLayer (layerId ()) ;
	pCopy->setLinetype (linetypeId ()) ;
	pCopy->setLinetypeScale (linetypeScale ()) ;
	pCopy->setLineWeight (lineWeight ()) ;
	pCopy->setColorIndex (colorIndex ()) ;
	pCopy->setVisibility (visibility ()) ;
	//- Pline
	double d, w0, w1 ;
	AcGePoint2d pt ;
	for ( unsigned int i =0 ; i < numVerts () ; i++ ) {
		getBulgeAt (i, d) ;
		getWidthsAt (i, w0, w1) ;
		getPointAt (i, pt) ;
		pCopy->addVertexAt (i, pt, d, w0, w1) ;
	}
	pCopy->setClosed (isClosed ()) ;
	if ( getConstantWidth (d) == Acad::eOk )
		pCopy->setConstantWidth (d) ;
	pCopy->setElevation (elevation ()) ;
	pCopy->setNormal (normal ()) ;
	pCopy->setPlinegen (hasPlinegen ()) ;
	pCopy->setThickness (thickness ()) ;

	TCHAR *pltName =plotStyleName () ;
	pCopy->setPlotStyleName (pltName, true) ;

	entitySet.append (pCopy) ;

	asdkOPLineStyle *pStyle ;
	if ( m_OPLineStyleId != AcDbObjectId::kNull ) {
		ARXOK ( acdbOpenObject (pStyle, m_OPLineStyleId, AcDb::kForRead) ) ;

		for ( Adesk::UInt32 i =0 ; i < pStyle->numOffsets () ; i++ ) {
			//- Create offset
			AcDbVoidPtrArray vpArr ;
			double off ;
			pStyle->offsetOf (i, off) ;
			if ( off == 0 )
				continue ;
			if ( AcDbPolyline::getOffsetCurves (off, vpArr) != Acad::eOk )
				continue ;

			for ( int j =0 ; j < vpArr.length () ; j++ ) {
				AcDbEntity *pCopy2 =AcDbEntity::cast ((AcRxObject *)vpArr [0]) ;
				//- Entity
				pCopy2->setLayer (layerId ()) ;
				pCopy2->setLinetypeScale (linetypeScale ()) ;
				pCopy2->setVisibility (visibility ()) ;
				//- Set properties based on style
				Adesk::UInt16 color ;
				AcDbObjectId ltId ;
				AcDb::LineWeight lwt ;
				pStyle->colorOf (i, color) ;
				pStyle->linetypeOf (i, ltId) ;
				pStyle->lineWeightOf (i, lwt) ;
				pCopy2->setColorIndex (color) ;
				pCopy2->setLinetype (ltId) ;
				pCopy2->setLineWeight (lwt) ;
				//- Polyline
				if ( pCopy2->isKindOf (AcDbPolyline::desc ()) )
					((AcDbPolyline *)pCopy2)->setElevation (elevation ()) ;

				pCopy2->setPlotStyleName (pltName, true) ;

				entitySet.append (pCopy2) ;
			}
		}
		pStyle->close () ;
	}
	delete [] pltName ;

	return (/*AcDbPolyline::subExplode (entitySet)*/Acad::eOk) ;
}