// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// Cyrille Fauvel
// Autodesk Developer Network

//-------------------------------------------------------------
//- asdkOPLineReactor.cpp : Implementation of asdkOPLineReactor
//-------------------------------------------------------------
#include "StdAfx.h"
#include "asdkOPLineReactor.h"
#include "asdkOPLineStyle.h"

//-------------------------------------------------------------
Adesk::UInt32 asdkOPLineReactor::kCurrentVersionNumber =1 ;

bool asdkOPLineReactor::m_bActive =true ;
AcArray<AcDbObjectId> asdkOPLineReactor::m_pLinesModified ;

//-----------------------------------------------------------------
//- When closing we will get an exception because a persistent reactor
//- will require to be re-open'ed. Here we are in read mode already, hence
//- it will fail. The following trick insures we can achieve our goal...
#define ARX_REACTOR_CLOSE(a,b); \
	{\
		AcDbObjectId idr =a->objectId () ;\
		a->close () ;\
		b.close () ;\
		AcDbObject *po =NULL ;\
		acdbOpenAcDbObject (po, idr, AcDb::kForRead, Adesk::kTrue) ;\
	}

//-------------------------------------------------------------
ACRX_DXF_DEFINE_MEMBERS (
	asdkOPLineReactor, AcDbObject,
	AcDb::kDHL_CURRENT, AcDb::kMReleaseCurrent, 
	AcDbProxyEntity::kNoOperation, ASDKOPLINEREACTOR,
ASDKOPLINEAPP
|Product Desc:     A description for your object
|Company:          Your company name
|WEB Address:      Your company WEB site address
)

//-------------------------------------------------------------
asdkOPLineReactor::asdkOPLineReactor () : AcDbObject () {
}

asdkOPLineReactor::~asdkOPLineReactor () {
}

//-------------------------------------------------------------
Acad::ErrorStatus asdkOPLineReactor::oPLineStyle (AcDbHardPointerId &OPLineStyle) {
	assertReadEnabled () ;
	OPLineStyle =m_OPLineStyle ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineReactor::setOPLineStyle (const AcDbHardPointerId OPLineStyle) {
	assertWriteEnabled () ;
	m_OPLineStyle =OPLineStyle ;
	return (Acad::eOk) ;
}

//-------------------------------------------------------------
//- AcDbObject protocols
//- Dwg Filing protocol
Acad::ErrorStatus asdkOPLineReactor::dwgOutFields (AcDbDwgFiler *pFiler) const {
	assertReadEnabled () ;
	//- Save parent class information first.
	Acad::ErrorStatus es =AcDbObject::dwgOutFields (pFiler) ;
	if ( es != Acad::eOk )
		return (es) ;
	//- Object version number needs to be saved first
	if ( (es =pFiler->writeUInt32 (asdkOPLineReactor::kCurrentVersionNumber)) != Acad::eOk )
		return (es) ;
	//- Output params
	//.....
	pFiler->writeHardPointerId (m_OPLineStyle) ;
	pFiler->writeUInt32 (m_entsToUpdate.length ()) ;
	for ( int i =0 ; i < m_entsToUpdate.length () ; i++ )
		pFiler->writeSoftPointerId (m_entsToUpdate.at (i)) ;

	return (pFiler->filerStatus ()) ;
}

Acad::ErrorStatus asdkOPLineReactor::dwgInFields (AcDbDwgFiler *pFiler) {
	assertWriteEnabled () ;
	//- Read parent class information first.
	Acad::ErrorStatus es =AcDbObject::dwgInFields (pFiler) ;
	if ( es != Acad::eOk )
		return (es) ;
	//- Object version number needs to be read first
	Adesk::UInt32 version =0 ;
	if ( (es =pFiler->readUInt32 (&version)) != Acad::eOk )
		return (es) ;
	if ( version > asdkOPLineReactor::kCurrentVersionNumber )
		return (Acad::eMakeMeProxy) ;
	//- Uncomment the 2 following lines if your current object implementation cannot
	//- support previous version of that object.
	//if ( version < asdkOPLineReactor::kCurrentVersionNumber )
	//	return (Acad::eMakeMeProxy) ;
	//- Read params
	//.....
	m_entsToUpdate.setLogicalLength (0) ;

	pFiler->readHardPointerId (&m_OPLineStyle) ;
	Adesk::UInt32 numOffsets ;
	pFiler->readUInt32 (&numOffsets) ;
	AcDbSoftPointerId eId ;
	for ( Adesk::UInt32 i =0 ; i < numOffsets ; i++ ) {
		pFiler->readSoftPointerId (&eId) ;
		m_entsToUpdate.append (eId) ;
	}

	return (pFiler->filerStatus ()) ;
}

Acad::ErrorStatus asdkOPLineReactor::subDeepClone (
	AcDbObject* pOwnerObject,
	AcDbObject*& pClonedObject,
	AcDbIdMapping& idMap,
	Adesk::Boolean isPrimary
) const {
	assertReadEnabled () ;
	Acad::ErrorStatus es =AcDbObject::subDeepClone (pOwnerObject, pClonedObject, idMap, isPrimary) ;

	switch ( idMap.deepCloneContext () ) {
		case AcDb::kDcInsert:
		case AcDb::kDcInsertCopy:
			//- We should force the copy of the style
			if ( !idMap.compute (AcDbIdPair (m_OPLineStyle, AcDbObjectId::kNull, true)) ) {
				//- Open the orignal dictionary object we want to clone
				AcDbObjectPointer<asdkOPLineStyle> pObj (m_OPLineStyle, AcDb::kForRead) ;
				//- Open the dictionary owner of the original dictionary object
				AcDbDictionaryPointer pDict (pObj->ownerId (), AcDb::kForRead) ;
				//- Clone the dictionary owner if necessary, if not open it for write
				AcDbDatabase *pDestDb =NULL ;
				idMap.destDb (pDestDb) ;
				AcDbDictionaryPointer pNOD (pDestDb->namedObjectsDictionaryId (), AcDb::kForRead) ;
				if ( pNOD->has (ASDK_OPLINESTYLE) ) {
					//- It is there
					AcDbObjectId idc ;
					pNOD->getAt (ASDK_OPLINESTYLE, idc) ;
					AcDbDictionaryPointer pClonedDict (idc, AcDb::kForWrite) ;
					//- Clone the dictionary object now
					asdkOPLineStyle *pClone =NULL ;
					pObj->deepClone (pClonedDict, (AcDbObject *&)pClone, idMap, true) ;
					//- Assign its owner ourself with an anonymous key
					pClonedDict->setAt (_T("*"), pClone, idc) ;
					//- We *must* close the clones ourself
					pClone->close () ;
				} else {
					//- Clone it
					pNOD->upgradeOpen () ;
					AcDbDictionary *pClonedDict =NULL ;
					pDict->deepClone (pNOD, (AcDbObject *&)pClonedDict, idMap, true) ;
					AcDbObjectId idc ;
					pNOD->setAt (ASDK_OPLINESTYLE, pClonedDict, idc) ;
					pNOD->downgradeOpen () ;
					//- We *must* close the clones ourself
					pClonedDict->close () ;
				}
			}
			break ;
		default:
			break ;
	}
	return (es) ;
}

//- Persistent reactor callbacks
//void asdkOPLineReactor::openedForModify (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::openedForModify (pDbObj) ;
//}
//
//void asdkOPLineReactor::cancelled (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::cancelled (pDbObj) ;
//}
//
//void asdkOPLineReactor::objectClosed (const AcDbObjectId objId) {
//	assertReadEnabled () ;
//	AcDbObject::objectClosed (objId) ;
//}
//
//void asdkOPLineReactor::goodbye (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::goodbye (pDbObj) ;
//}
//
//void asdkOPLineReactor::copied (const AcDbObject *pDbObj, const AcDbObject *pNewObj) {
//	assertReadEnabled () ;
//	AcDbObject::copied (pDbObj, pNewObj) ;
//}

void asdkOPLineReactor::erased (const AcDbObject *pDbObj, Adesk::Boolean bErasing) {
	assertNotifyEnabled () ;
	if ( asdkOPLineReactor::m_bActive == true ) {
		//- First: Is it the main entity being erased or not?
		AcDbDictionaryPointer pExtDct (ownerId (), AcDb::kForRead) ;
		if ( pExtDct.openStatus () != Acad::eOk )
			return ;
		AcDbObjectId mainEntId =pExtDct->ownerId () ;

		if ( mainEntId != pDbObj->objectId () ) {
			AcDbObjectPointer<AcDbObject> p (mainEntId, AcDb::kForWrite) ;
			if ( p.openStatus () == Acad::eOk ) {
				p->erase (bErasing) ;
				ARX_REACTOR_CLOSE (this, p) ;
			}
		}

		for ( int i =0 ; i < m_entsToUpdate.length () ; i++ ) {
			AcDbObjectPointer<AcDbObject> p (m_entsToUpdate.at (i), AcDb::kForWrite) ;
			if ( p.openStatus () != Acad::eOk)
				continue ;
			p->erase (bErasing) ;
			ARX_REACTOR_CLOSE (this, p) ;
		}
	}
	AcDbObject::erased (pDbObj, bErasing) ;
}

void asdkOPLineReactor::modified (const AcDbObject *pDbObj) {
	assertNotifyEnabled () ;
	//- We cannot do the following... We *must* wait that the command is ending...
#ifdef MAKE_IT_THERE
	if ( AsdkOPLineReactor::m_bActive == true ) {
		//- First: Is it the main entity being modified or not?
		AcDbDictionaryPointer pExtDct (ownerId (), AcDb::kForRead, Adesk::kTrue) ;
		if ( pExtDct.openStatus () != Acad::eOk )
			return ;
		AcDbObjectId mainEntId ;
		mainEntId =pExtDct->ownerId () ;

		if ( mainEntId == dbObj->objectId () ) {
			//- Yes it is! So cool...
			MakeOPLine (dbObj) ;
		} else {
			//- No it isn't, the notification is coming from one of the counter-partner
			ModifyOPLineFrom (mainEntId, dbObj) ;
		}
	}
#endif
	//- Instead we just store the Id of modified entities, and wait for commandEnded()
	//- NB: We could do the same for erased, but for erased it is working that way too...
	if ( asdkOPLineReactor::m_bActive == true ) {
		if ( !asdkOPLineReactor::m_pLinesModified.contains (this->objectId ()) ) {
			//- Always by pair...
			asdkOPLineReactor::m_pLinesModified.append (this->objectId ()) ;
			asdkOPLineReactor::m_pLinesModified.append (pDbObj->objectId ()) ;
		}
	}
	AcDbObject::modified (pDbObj) ;
}

//void asdkOPLineReactor::modifiedGraphics (const AcDbEntity *pDbEnt) {
//	assertReadEnabled () ;
//	AcDbObject::modifiedGraphics (pDbEnt) ;
//}
//
//void asdkOPLineReactor::modifiedXData (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::modifiedXData (pDbObj) ;
//}
//
//void asdkOPLineReactor::subObjModified (const AcDbObject *pMainbObj, const AcDbObject *pSubObj) {
//	assertReadEnabled () ;
//	AcDbObject::subObjModified (pMainbObj, pSubObj) ;
//}
//
//void asdkOPLineReactor::modifyUndone (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::modifyUndone (pDbObj) ;
//}
//
//void asdkOPLineReactor::reappended (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::reappended (pDbObj) ;
//}
//
//void asdkOPLineReactor::unappended (const AcDbObject *pDbObj) {
//	assertReadEnabled () ;
//	AcDbObject::unappended (pDbObj) ;
//}
//

//---------------------------------------------------------------------
AcString asdkOPLineReactor::printIt () {
	AcDbDictionaryPointer pExtDct (ownerId (), AcDb::kForRead);
	if ( pExtDct.openStatus () != Acad::eOk )
		return (_T("")) ;
	AcDbObjectId mainEntId =pExtDct->ownerId () ;

	AcString st ;
	st.format (
		_T("\nReactor: %ld\n	Style: %ld\n	Main entity link: %ld\n	Entity(ies) to update: "),
		objectId ().asOldId (),
		m_OPLineStyle.asOldId (),
		mainEntId.asOldId ()
	) ;
	AcString temp ;
	for ( int i =0 ; i < m_entsToUpdate.length () ; i++ ) {
		temp.format (_T(" %ld"), m_entsToUpdate.at (i).asOldId ()) ;
		st +=temp ;
	}
	st +=_T("\n") ;

	return (st) ;
}

//---------------------------------------------------------------------
void asdkOPLineReactor::MakeOPLine (const AcDbObject *pObj) {
	bool b =asdkOPLineReactor::m_bActive ;
	asdkOPLineReactor::m_bActive =false ;

	AcDbPolyline *pLine =AcDbPolyline::cast (pObj) ;
	if ( pLine == NULL ) {
		asdkOPLineReactor::m_bActive =b ;
		return ;
	}
	//- Iterate through all of the offsets defined in m_OPLineStyle
	AcDbObjectPointer<asdkOPLineStyle> pStyle (m_OPLineStyle, AcDb::kForRead, Adesk::kTrue) ;
	if ( pStyle.openStatus () != Acad::eOk ) {
		asdkOPLineReactor::m_bActive =b ;
		return ;
	}
	//- Update (and create) pline(s) for the OPLine
	AcDbObjectPointer<AcDbPolyline> pLine2 ;
	try {
		for ( int i =0 ; i < (int)pStyle->numOffsets () ; i++ ) {
			//- Create offset
			double off =0 ;
			pStyle->offsetOf (i, off) ;
			if ( off == 0 )
				continue ;
			AcDbVoidPtrArray vpArr ;
			if (   pLine->getOffsetCurves (off, vpArr) != Acad::eOk
				|| vpArr.length () == 0
			)
				continue ;
			for ( int j =1 ; j < vpArr.length () ; j++ )
				delete ((AcDbObject *)vpArr.at (j)) ;
			vpArr.setLogicalLength (1) ;

			AcDbPolyline *pCopy =AcDbPolyline::cast ((AcRxObject *)vpArr [0]) ;
			if ( pCopy == NULL ) {
				delete ((AcDbObject *)vpArr.at (0)) ;
				continue ;
			}
			//- Get an offset pline object
			AcDbObjectId id =(i < m_entsToUpdate.length () ? m_entsToUpdate.at (i) : AcDbObjectId::kNull) ;
			if ( id == AcDbObjectId::kNull ) {
				pLine2.create () ;
				pLine2->setDatabaseDefaults (database ()) ;
				AcDbBlockTableRecordPointer pRec (pObj->ownerId (), AcDb::kForWrite, Adesk::kTrue) ;
				pRec->appendAcDbEntity (id, pLine2) ;
				//- Reactor assignement
				pLine2->addPersistentReactor (this->objectId ()) ;
				m_entsToUpdate.append (id) ;
			} else {
				pLine2.open (id, AcDb::kForWrite, Adesk::kTrue) ;
				if ( pLine2.openStatus () != Acad::eOk ) {
					delete ((AcDbObject *)vpArr.at (0)) ;
					continue ;
				}
			}
			//- Reset and properties/vertex assignments
			pLine2->reset (false, 0) ;
			double d, w0, w1 ;
			AcGePoint2d pt ;
			for ( unsigned int k =0 ; k < pCopy->numVerts () ; k++ ) {
				pCopy->getBulgeAt (k, d) ;
				pCopy->getWidthsAt (k, w0, w1) ;
				pCopy->getPointAt (k, pt) ;
				pLine2->addVertexAt (k, pt, d, w0, w1) ;
			}
			pLine2->setClosed (pCopy->isClosed ()) ;
			if ( pCopy->getConstantWidth (d) == Acad::eOk )
				pLine2->setConstantWidth (d) ;
			pLine2->setPlinegen (pCopy->hasPlinegen ()) ;
			delete pCopy ;
			//- Set properties based on style
			Adesk::UInt16 color ;
			AcDbObjectId ltId ;
			AcDb::LineWeight lwt ;
			pStyle->colorOf (i, color) ;
			pStyle->lineWeightOf (i, lwt) ;
			pLine2->setColorIndex (color) ;
			pLine2->setLineWeight (lwt) ;

			ARX_REACTOR_CLOSE (this, pLine2) ;
		}
		//- Erase unused pline(s)
		for ( int i =m_entsToUpdate.length () - 1 ; i > (int)pStyle->numOffsets () - 1 ; i-- ) {
			AcDbObjectPointer<AcDbPolyline> pLine2 (m_entsToUpdate.at (i), AcDb::kForWrite, Adesk::kTrue) ;
			if ( pLine2.openStatus () != Acad::eOk )
				continue ;
			pLine2->removePersistentReactor (this->objectId ()) ;
			this->upgradeOpen () ;
			m_entsToUpdate.removeAt (i) ;
			if ( !pLine2->isErased () )
				pLine2->erase (true) ;
			ARX_REACTOR_CLOSE (this, pLine2) ;
		}
		m_entsToUpdate.setLogicalLength (pStyle->numOffsets ()) ;
	} catch (...) {
		if ( pLine2.openStatus () != Acad::eOk ) {
			m_pLinesModified.append (this->objectId ()) ;
			m_pLinesModified.append (m_entsToUpdate.at (0)) ;
		}
	}

	asdkOPLineReactor::m_bActive =b ;
}

void asdkOPLineReactor::ModifyOPLineFrom (AcDbObjectId mainId, const AcDbObject *pObj) {
	bool b =asdkOPLineReactor::m_bActive ;
	asdkOPLineReactor::m_bActive =false ;

	AcDbObjectPointer<AcDbPolyline> pLineMain (mainId, AcDb::kForWrite, Adesk::kTrue) ;
	if ( pLineMain.openStatus () != Acad::eOk ) {
		//- Main entity already opened! It means the main entity is modifying all
		//- its offset'ed pline(s). To not fo into an infinite loop, just return...
		asdkOPLineReactor::m_bActive =b ;
		return ;
	}

	int i =0 ;
	AcDbPolyline *pLine =AcDbPolyline::cast (pObj) ;
	if ( pLine == NULL || m_entsToUpdate.find (pObj->objectId (), i, 0) == false ) {
		ARX_REACTOR_CLOSE (this, pLineMain) ;
		asdkOPLineReactor::m_bActive =b ;
		return ;
	}

	AcDbObjectPointer<asdkOPLineStyle> pStyle (m_OPLineStyle, AcDb::kForRead, Adesk::kTrue) ;
	if ( pStyle.openStatus () != Acad::eOk ) {
		ARX_REACTOR_CLOSE (this, pLineMain) ;
		asdkOPLineReactor::m_bActive =b ;
		return ;
	}
	double off =0 ;
	pStyle->offsetOf (i, off) ;
	if ( off == 0 ) {
		ARX_REACTOR_CLOSE (this, pLineMain) ;
		asdkOPLineReactor::m_bActive =b ;
		return ;
	}

	try {
		AcDbVoidPtrArray vpArr ;
		if (   pLine->getOffsetCurves (-off, vpArr) != Acad::eOk
			|| vpArr.length () == 0
		) {
			ARX_REACTOR_CLOSE (this, pLineMain) ;
			asdkOPLineReactor::m_bActive =b ;
			return ;
		}
		for ( int j =1 ; j < vpArr.length () ; j++ )
			delete vpArr [j] ;
		vpArr.setLogicalLength (1) ;

		AcDbPolyline *pCopy =AcDbPolyline::cast ((AcRxObject *)vpArr [0]) ;
		if ( pCopy == NULL ) {
			delete vpArr [0] ;
			ARX_REACTOR_CLOSE (this, pLineMain) ;
			asdkOPLineReactor::m_bActive =b ;
			return ;
		}

		pLineMain->reset (false, 0) ;

		double d, w0, w1 ;
		AcGePoint2d pt ;
		for ( int i =0 ; i < (int)pCopy->numVerts () ; i++ ) {
			pCopy->getBulgeAt (i, d) ;
			pCopy->getWidthsAt (i, w0, w1) ;
			pCopy->getPointAt (i, pt) ;
			pLineMain->addVertexAt (i, pt, d, w0, w1) ;
		}
		pLineMain->setClosed (pCopy->isClosed ()) ;
		if ( pCopy->getConstantWidth (d) == Acad::eOk )
			pLineMain->setConstantWidth (d) ;
		pLineMain->setPlinegen (pCopy->hasPlinegen ()) ;
		delete pCopy ;
	} catch (...) {
		m_pLinesModified.append (this->objectId ()) ;
		m_pLinesModified.append (mainId) ;
	}
	ARX_REACTOR_CLOSE (this, pLineMain) ;
	asdkOPLineReactor::m_bActive =b ;
}

//-------------------------------------------------------------
void asdkOPLineReactor::modify (const AcDbObject *pObj) {
	//- First: Is it the main entity being modified or not?
	AcDbDictionaryPointer pExtDct (ownerId (), AcDb::kForRead, Adesk::kTrue) ;
	if ( pExtDct.openStatus () != Acad::eOk )
		return ;
	AcDbObjectId mainEntId =pExtDct->ownerId () ;
	if ( mainEntId == pObj->objectId () ) {
		//- Yes it is! So cool...
		MakeOPLine (pObj) ;
	} else {
		//- No it isn't, the notification is coming from one of the counter-partner
		ModifyOPLineFrom (mainEntId, pObj) ;
		AcDbObjectPointer<AcDbObject> pObj2 (mainEntId, AcDb::kForNotify, Adesk::kTrue) ;
		if ( pObj2.openStatus () == Acad::eOk ) {
			modify (pObj2) ;
		}
	}
}