// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// Cyrille Fauvel
// Autodesk Developer Network

//-------------------------------------------------------------------------
//- asdkOPLineStyle.cpp : Implementation of asdkOPLineStyle
//-------------------------------------------------------------------------
#include "StdAfx.h"
#include "asdkOPLineStyle.h"

//-------------------------------------------------------------------------
Adesk::UInt32 asdkOPLineStyle::kCurrentVersionNumber =1 ;
TCHAR kMchr =_T('$') ; //- MDI safe

typedef Acad::ErrorStatus (*fct_acedGetCurDwgXrefGraph)(AcDbXrefGraph&, Adesk::Boolean includeGhosts) ;
bool wblockObjectsUnmangleName (AcString &csName) ;
const TCHAR *xrefName (const AcDbDatabase *pDb) ;

//-------------------------------------------------------------------------
ACRX_DXF_DEFINE_MEMBERS (
	asdkOPLineStyle, AcDbObject,
	AcDb::kDHL_CURRENT, AcDb::kMReleaseCurrent, 
	AcDbProxyEntity::kNoOperation, ASDKOPLINESTYLE,
ASDKOPLINEAPP
|Product Desc:     A description for your object
|Company:          Your company name
|WEB Address:      Your company WEB site address
)

//-------------------------------------------------------------------------
asdkOPLineStyle::asdkOPLineStyle () : AcDbObject () {
	m_iNumOffsets =0 ;
	m_offsetArray.setLogicalLength (0) ;
	m_ltypeArray.setLogicalLength (0) ;
	m_colorArray.setLogicalLength (0) ;
	m_lweightArray.setLogicalLength (0) ;
}

asdkOPLineStyle::~asdkOPLineStyle () {
}

//-------------------------------------------------------------------------
//- AcDbObject protocols
//- Dwg Filing protocol
Acad::ErrorStatus asdkOPLineStyle::dwgOutFields (AcDbDwgFiler *pFiler) const {
	assertReadEnabled () ;
	//- Save parent class information first.
	Acad::ErrorStatus es =AcDbObject::dwgOutFields (pFiler) ;
	if ( es != Acad::eOk )
		return (es) ;
	//- Object version number needs to be saved first
	if ( (es =pFiler->writeUInt32 (asdkOPLineStyle::kCurrentVersionNumber)) != Acad::eOk )
		return (es) ;
	//- Output params
	//.....
	pFiler->writeUInt32 (m_iNumOffsets) ;
	for ( Adesk::UInt32 i =0 ; i < m_iNumOffsets ; i++ ) {
		pFiler->writeDouble (m_offsetArray [i]) ;
		pFiler->writeHardPointerId (m_ltypeArray [i]) ;
		pFiler->writeUInt16 (m_colorArray [i]) ;
		pFiler->writeUInt16 ((Adesk::UInt16)m_lweightArray [i]) ;
	}

	return (pFiler->filerStatus ()) ;
}

Acad::ErrorStatus asdkOPLineStyle::dwgInFields (AcDbDwgFiler *pFiler) {
	assertWriteEnabled () ;
	//- Read parent class information first.
	Acad::ErrorStatus es =AcDbObject::dwgInFields (pFiler) ;
	if ( es != Acad::eOk )
		return (es) ;
	//- Object version number needs to be read first
	Adesk::UInt32 version =0 ;
	if ( (es =pFiler->readUInt32 (&version)) != Acad::eOk )
		return (es) ;
	if ( version > asdkOPLineStyle::kCurrentVersionNumber )
		return (Acad::eMakeMeProxy) ;
	//- Uncomment the 2 following lines if your current object implementation cannot
	//- support previous version of that object.
	//if ( version < asdkOPLineStyle::kCurrentVersionNumber )
	//	return (Acad::eMakeMeProxy) ;
	//- Read params
	//.....
	m_iNumOffsets =0 ;
	m_offsetArray.setLogicalLength (0) ;
	m_ltypeArray.setLogicalLength (0) ;
	m_colorArray.setLogicalLength (0) ;
	m_lweightArray.setLogicalLength (0) ;

	pFiler->readUInt32 (&m_iNumOffsets) ;
	AcDbHardPointerId eId ;
	double offset ;
	Adesk::UInt16 color ;
	Adesk::UInt16 weight ;
	for ( Adesk::UInt32 i =0 ; i < m_iNumOffsets ; i++ ) {
		pFiler->readDouble (&offset) ;
		m_offsetArray.append (offset) ;
		pFiler->readHardPointerId (&eId) ;
		m_ltypeArray.append (eId) ;
		pFiler->readUInt16 (&color) ;
		m_colorArray.append (color) ;
		pFiler->readUInt16 (&weight) ;
		m_lweightArray.append ((AcDb::LineWeight)(weight)) ;
	}

	return (pFiler->filerStatus ()) ;
}

//-------------------------------------------------------------------------
bool asdkOPLineStyle::isEquivalentDuplicate (AcDbObjectId otherId, AcDb::DuplicateRecordCloning drc) const {
	//- It doesn't make sense to call this function unless using
	//- one of the following duplicate handling types.  The function
	//- probably would not work properly, either.
	assert ( drc == AcDb::kDrcIgnore || drc == AcDb::kDrcReplace || drc == AcDb::kDrcUnmangleName ) ;

	//- If you want the 'Standard' style being always equivalent
	//- then activate the following code
	/*AcDbDictionaryPointer pDct (ownerId (), AcDb::kForRead) ;
	TCHAR *name =NULL ;
	pDct->nameAt (objectId (), name) ;
	AcString mName (name);
	delete [] name ;

	if ( mName != ASDK_OPLINESTYLE_STD )
		return (true) ;*/

	AcDbObjectPointer<asdkOPLineStyle> pStyle (otherId, AcDb::kForRead) ;
	if ( pStyle.openStatus () != Acad::eOk )
		return (false) ;

	if ( m_iNumOffsets != pStyle->numOffsets () )
		return (false) ;

	bool bSame =true ;
	for ( Adesk::UInt32 i =0 ; i < m_iNumOffsets ; i++ ) {
		if (   m_colorArray.at (i) != pStyle->m_colorArray.at (i)
			|| m_lweightArray.at (i) != pStyle->m_lweightArray.at (i)
			|| m_offsetArray.at (i) != pStyle->m_offsetArray.at (i)
		) {
			bSame =false ;
			break ;
		}
		//- Linetypes are more difficult. The Ids will not compare,
		//- so we have to compare by name - including doing any
		//- unmangling, if that is the Drc mode.
		TCHAR *name =NULL ;
		AcDbSymUtil::getSymbolName (name, m_ltypeArray.at (i)) ;
		AcString csMyLtName (name) ;
		delete [] name ;
		AcDbSymUtil::getSymbolName (name, pStyle->m_ltypeArray.at (i)) ;
		AcString csOtherLtName (name) ;
		delete [] name ;
		if ( drc == AcDb::kDrcUnmangleName )
			wblockObjectsUnmangleName (csMyLtName) ;
		if ( acdbSymUtil ()->compareSymbolName (csMyLtName, csOtherLtName) != 0 ) {
			bSame =false ;
			break ;
		}
	}

	return (bSame) ;
}

Acad::ErrorStatus asdkOPLineStyle::subWblockClone (
	AcRxObject *pOwnerObject,
	AcDbObject *&pClonedObject,
	AcDbIdMapping& idMap,
	Adesk::Boolean isPrimary
) const {
	assertReadEnabled () ;
	Acad::ErrorStatus es ;
	AcDb::DeepCloneType dc =idMap.deepCloneContext () ;

	if (   dc != AcDb::kDcXrefBind
		&& dc != AcDb::kDcXrefInsert
		&& dc != AcDb::kDcWblkObjects
	)
		return (AcDbObject::subWblockClone (pOwnerObject, pClonedObject, idMap, isPrimary)) ;

	pClonedObject =NULL ;

	//- Make sure we have not already done this one.
	AcDbIdPair idPair (objectId (), AcDbObjectId::kNull, true) ;
	if ( idMap.compute (idPair) && (idPair.value () != AcDbObjectId::kNull) )
		return (Acad::eOk) ;

	//- Xref Bind will always pass in an AcDbDatabase* for pOwner, but
	//- kDcWblkObjects may also pass in the destination dictionary already
	//- open. So, if pDestDb == NULL, we know that we did not open the
	//- dictionary here.
	AcDbDictionary *pDestDict =NULL ;
	AcDbDatabase *pDestDb =AcDbDatabase::cast (pOwnerObject) ;
	if ( pDestDb ) {
		AcDbDictionary *pNOD =NULL ;
		pDestDb->getNamedObjectsDictionary (pNOD, AcDb::kForRead) ;
		es =pNOD->getAt (ASDK_OPLINESTYLE, (AcDbObject *&)pDestDict, AcDb::kForWrite) ;
		pNOD->close () ;
		if ( es == Acad::eKeyNotFound ) {
			//- It is safe to clone as usual in 2000
			return (AcDbObject::subWblockClone (pOwnerObject, pClonedObject, idMap, isPrimary)) ;
		}
		if ( es != Acad::eOk )
			return (es) ;
	} else {
		pDestDict =AcDbDictionary::cast (pOwnerObject) ;
	}
	assert ( pDestDict != NULL ) ;
	assert ( pDestDb != NULL || dc == AcDb::kDcWblkObjects ) ;

	//- Now determine how we should handle duplicates. Normally
	//- we would honor the DuplicateRecordCloning type, but OPLines
	//- cannot handle swapping styles, so we never ignore or replace
	//- the records.
	AcDb::DuplicateRecordCloning drc =idMap.duplicateRecordCloning();

	//- We never replace OPLineStyles. Since we only ignore exact
	//- duplicates, replace and ignore are identical anyway.
	if ( drc == AcDb::kDrcReplace )
		drc =AcDb::kDrcIgnore ;

	AcDbDictionaryPointer pDct (ownerId (), AcDb::kForRead) ;
	TCHAR *name =NULL ;
	pDct->nameAt (objectId (), name) ;
	AcString csName (name) ;
	delete [] name ;

	bool bNewName =false ;
	AcDbObjectId dupId ;
	int idx =0 ; //- mangle index
	AcString csRootName, csXrefName ;

	switch ( drc ) {
		case AcDb::kDrcUnmangleName:
			bNewName =wblockObjectsUnmangleName (csName) ;
			//- fall through ...
		case AcDb::kDrcIgnore:
			if ( pDestDict->getAt (csName, dupId) == Acad::eOk ) {
				//- Normally, we consider a duplicate name enough to
				//- ignore - but if the style is really not a duplicate,
				//- OPLines crash - so we have to further verify the 
				//- duplicate against the entire OPLineStyle.
				if ( isEquivalentDuplicate (dupId, drc) ) {
					if ( pDestDb )
						pDestDict->close () ;
					idPair.setKey (objectId ()) ;
					idPair.setValue (dupId) ;
					idPair.setIsCloned (false) ;
					return (idMap.assign (idPair)) ;
				} else {
					//- Use an alternate mangling 
					csRootName =csName ;
					do {
						csName.format (_T("%s%d"), csRootName, ++idx) ;
						assert (idx < 1000) ;
					} while ( pDestDict->has (csName) ) ;
					bNewName = true;
				} 
			}
			break ;
		case AcDb::kDrcMangleName:
			//- Use $0$name mangling
			csRootName =csName ;
			do {
				csName.format (_T("%c%d%c%s"), kMchr, idx++, kMchr, csRootName) ;
				assert ( idx < 1000 ) ;
			} while ( pDestDict->has (csName) ) ;
			bNewName =true ;
			break ;
		case AcDb::kDrcXrefMangleName:
			//- Use xref$0$name mangling

			//- Get the xref name from which this style comes from. We need
			//- to look this up, because unlike STRs, OPLineStyles are not
			//- attached, so we don't start out with a name of the form,
			//- "<xref>|<name>".  The bindManager uses graph of the current
			//- drawing's Xrefs to find the correct name. If the name comes
			//- back as NULL, then this OPLineStyle is not from an Xref'ed
			//- database - and should not have gotten this call in the 
			//- first place.

			//- BindManager is only an issue in AutoCAD, which has XREF-BIND
			//- Command. In DWG U, we should only get here via an API, not an
			//- XREF-BIND operation, which means we shouldn't ever enter this
			//- case in DWG U.
	        if ( acdbHostApplicationServices ()->acadInternalServices () ) {
				csXrefName =xrefName (database ()) ;
				assert ( !csXrefName.isEmpty ()) ;
				csRootName =csName ;
				do {
					csName.format (_T("%s%c%d%c%s"), csXrefName, kMchr, idx++, kMchr, csRootName) ;
					assert(idx < 1000);
				} while ( pDestDict->has (csName) ) ;
				bNewName =true ;
			}
			break ;
		default:  
			assert ( 0 ) ;
	}

	//- We pass in pDict so that wblockClone knows that we are going
	//- to call setAt(), and it will not set isOwnerXlated to false.
	//- Since we know the right owner and will do the setAt() here, the
	//- setting of isPrimary does not matter. Also, we can get away
	//- with calling setAt() here because it is not possible to nest
	//- these calls. In other words, there is no way an OPLineStyle
	//- entry can reference anything which would in turn reference
	//- another OPLineStyle. 
	es =AcDbObject::subWblockClone (pDestDict, pClonedObject, idMap, isPrimary) ;
	if ( es != Acad::eOk || pClonedObject == NULL ) {
		if ( pDestDb )
			pDestDict->close () ;
		return (es) ;
	}

	AcDbObjectId id ;
	es =pDestDict->setAt (csName, pClonedObject, id) ;

	//- if OPLineStyles have duplicate name placements, both in the
	//- dictionary, and in their own records. Do this:
	//if ( bNewName && (es == Acad::eOk) )
	//	es =((AsdkOPLineStyle *)pClonedObject)->setName (csName) ;

	if ( pDestDb )
		pDestDict->close () ;

	//- If no errors, leave the clone open, as we are still
	//- inside wblockClone().
	if ( es != Acad::eOk ) {
		pClonedObject->close () ;
		pClonedObject =NULL ;
	}
	return (es) ;
}

//-------------------------------------------------------------------------
Acad::ErrorStatus asdkOPLineStyle::setOffsetOf (Adesk::UInt32 index, double off) {
	assertWriteEnabled () ;
	if ( (index >= m_iNumOffsets) || (index < 0) )
		return (Acad::eInvalidIndex) ;
	m_offsetArray [index] =off ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::offsetOf (Adesk::UInt32 index, double &off) {
	assertReadEnabled () ;
	if ( (index >= m_iNumOffsets) || (index < 0) )
		return (Acad::eInvalidIndex) ;
	off =m_offsetArray [index] ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::appendOffset (double off) {
	//- These are the defaults
	AcDbObjectId stdId ;
	//- Get the objectId of continuous linetype
	AcDbDatabase *pDb =database () ;
	if ( pDb == NULL )
		pDb =acdbHostApplicationServices ()->workingDatabase () ;
	Acad::ErrorStatus es =AcDbSymbolUtilities::getLinetypeId (stdId, _T("BYLAYER"), pDb) ;
	assert( es == Acad::eOk ) ;
	return (appendOffset (off, stdId, 256, AcDb::kLnWt000)) ;
}

Acad::ErrorStatus asdkOPLineStyle::appendOffset (double off, AcDbObjectId ltId, Adesk::UInt16 color, AcDb::LineWeight lwt) {
	assertWriteEnabled () ;
	m_iNumOffsets++ ;
	m_offsetArray.append (off) ;
	m_ltypeArray.append (ltId) ;
	m_colorArray.append (color) ;
	m_lweightArray.append (lwt) ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::deleteOffset (Adesk::UInt32 index) {
	assertWriteEnabled () ;
	if ( index >= m_iNumOffsets )
		return (Acad::eInvalidIndex) ;
	m_offsetArray.removeAt (index) ;
	m_ltypeArray.removeAt (index) ;
	m_colorArray.removeAt (index) ;
	m_lweightArray.removeAt (index) ;
	m_iNumOffsets-- ;
	return (Acad::eOk) ;
}

Adesk::UInt32 asdkOPLineStyle::numOffsets () {
	assertReadEnabled () ;
	return (m_iNumOffsets) ;
}

Acad::ErrorStatus asdkOPLineStyle::setLinetypeOf (Adesk::UInt32 index, AcDbObjectId eId) {
	assertWriteEnabled () ;
	if ( (index >= m_iNumOffsets) || (index < 0) )
		return (Acad::eInvalidIndex) ;
	if ( eId == AcDbObjectId::kNull )
		return (Acad::eNullObjectId) ;
	m_ltypeArray [index] =eId ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::linetypeOf (Adesk::UInt32 index, AcDbObjectId &eId) {
	assertReadEnabled () ;
	if ( index >= m_iNumOffsets )
		return (Acad::eInvalidIndex) ;
	eId =m_ltypeArray [index] ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::setColorOf (Adesk::UInt32 index, Adesk::UInt16 color) {
	assertWriteEnabled () ;
	if ( index >= m_iNumOffsets )
		return (Acad::eInvalidIndex) ;
	if ( (color > 256) || (color < 0) )
		return (Acad::eInvalidInput) ;
	m_colorArray [index] =color ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::colorOf (Adesk::UInt32 index, Adesk::UInt16 &color) {
	assertReadEnabled () ;
	if ( index >= m_iNumOffsets )
		return (Acad::eInvalidIndex) ;
	color =m_colorArray [index] ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::setLineWeightOf (Adesk::UInt32 index, AcDb::LineWeight lw) {
	assertWriteEnabled () ;
	AcDbDatabase *pDb =database () ;
	if ( pDb == NULL )
		pDb =acdbHostApplicationServices ()->workingDatabase () ;
	if ( index >= m_iNumOffsets ) 
		return (Acad::eInvalidIndex) ;
	if ( !pDb->isValidLineWeight (lw) )
		return (Acad::eInvalidInput) ;
	m_lweightArray [index] =lw ;
	return (Acad::eOk) ;
}

Acad::ErrorStatus asdkOPLineStyle::lineWeightOf (Adesk::UInt32 index, AcDb::LineWeight &lw) {
	assertReadEnabled () ;
	if ( index >= m_iNumOffsets )
		return (Acad::eInvalidIndex) ;
	lw =m_lweightArray [index] ;
	return (Acad::eOk) ;
}

//- Utilities
//- Does the unmangling specific to AcDb::kDrcUnmangleName, which
//- removes the $0$ from the beginning of the name. Any other
//- mangling, or usage of $ in a name is ignored. If the name
//- does not start with $0$, it is not changed. Returns true
//- if the name was unmangled.
bool wblockObjectsUnmangleName (AcString &csName) {
	int namePos ;
	int nameLen ;

	if ( csName.kACharPtr () [0] == kMchr ) {
		nameLen =csName.length () ;
		bool bTruncate =false ;
		bool bFoundNumber =false ;
		for ( namePos =1 ; namePos < nameLen ; namePos++ ) {
			if (   csName.kACharPtr () [namePos] >= _T('0')
				&& csName.kACharPtr () [namePos] <= _T('9')
			) {
				bFoundNumber =true ;
			} else {
				if ( bFoundNumber && csName.kACharPtr () [namePos] == kMchr )
					bTruncate =true ;
				break ;
			}
		}

		//- Only do it if there is a name to the right 
		//- of the $0$.
		if ( bTruncate && (++namePos < nameLen) ) {
			csName =csName.substrRev (nameLen - namePos) ; //.Right (nameLen - namePos) ;
			return (true) ;
		}
	}
	return (false) ;
}

const TCHAR *xrefName (const AcDbDatabase *pDb) {
	AcDbXrefGraph graph ;
	//acedGetCurDwgXrefGraph (graph) ;
	fct_acedGetCurDwgXrefGraph fct ;
	fct =(fct_acedGetCurDwgXrefGraph)::GetProcAddress (
		::GetModuleHandle (_T("acad.exe")),
		"?acedGetCurDwgXrefGraph@@YA?AW4ErrorStatus@Acad@@AEAVAcDbXrefGraph@@H@Z"
	) ;
	if ( fct == NULL )
		return (_T("")) ;
	(*fct) (graph, false) ;
	AcDbXrefGraphNode *pNode =graph.xrefNode (pDb) ;
	if ( pNode == NULL ) {
		assert ( pNode != NULL ) ;
		return (NULL) ;
	}
	return (pNode->name ()) ;
}