// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// Cyrille Fauvel
// Autodesk Developer Network

//-----------------------------------------------------------------------------
//----- asdkOPLineStyle.h : Declaration of the asdkOPLineStyle
//-----------------------------------------------------------------------------
#pragma once

#ifdef OPLINE_MODULE
#define DLLIMPEXP __declspec(dllexport)
#else
//----- Note: we don't use __declspec(dllimport) here, because of the
//----- "local vtable" problem with msvc. If you use __declspec(dllimport),
//----- then, when a client dll does a new on the class, the object's
//----- vtable pointer points to a vtable allocated in that client
//----- dll. If the client dll then passes the object to another dll,
//----- and the client dll is then unloaded, the vtable becomes invalid
//----- and any virtual calls on the object will access invalid memory.
//-----
//----- By not using __declspec(dllimport), we guarantee that the
//----- vtable is allocated in the server dll during the ctor and the
//----- client dll does not overwrite the vtable pointer after calling
//----- the ctor. And, since we expect the server dll to remain in
//----- memory indefinitely, there is no problem with vtables unexpectedly
//----- going away.
#define DLLIMPEXP
#endif

//-----------------------------------------------------------------------------
#include "dbmain.h"

#define ASDK_OPLINESTYLE		_T("Asdk_OPLineStyle")
#define ASDK_OPLINESTYLE_STD	_T("Standard")

//-----------------------------------------------------------------------------
class DLLIMPEXP asdkOPLineStyle : public AcDbObject {
	Adesk::UInt32 m_iNumOffsets ;
	AcArray<double> m_offsetArray ;
	AcDbObjectIdArray m_ltypeArray ;
	AcArray<Adesk::UInt16> m_colorArray ;
	AcArray<AcDb::LineWeight> m_lweightArray ;

public:
	ACRX_DECLARE_MEMBERS(asdkOPLineStyle) ;

protected:
	static Adesk::UInt32 kCurrentVersionNumber ;

public:
	asdkOPLineStyle () ;
	virtual ~asdkOPLineStyle () ;

	//----- AcDbObject protocols
	//- Dwg Filing protocol
	virtual Acad::ErrorStatus dwgOutFields (AcDbDwgFiler *pFiler) const ;
	virtual Acad::ErrorStatus dwgInFields (AcDbDwgFiler *pFiler) ;

	bool isEquivalentDuplicate (AcDbObjectId otherId, AcDb::DuplicateRecordCloning drc) const ;
	virtual Acad::ErrorStatus subWblockClone (
		AcRxObject *pOwnerObject,
		AcDbObject *&pClonedObject,
		AcDbIdMapping &idMap,
		Adesk::Boolean isPrimary =true
	) const ;

	virtual Acad::ErrorStatus setStyleFrom (AcDbObjectId sourceId) { return (Acad::eNotImplementedYet) ; }

	virtual Acad::ErrorStatus setOffsetOf (Adesk::UInt32 index, double off) ;
	virtual Acad::ErrorStatus offsetOf (Adesk::UInt32 index, double &off) ;
	virtual Acad::ErrorStatus appendOffset (double off) ;
	virtual Acad::ErrorStatus appendOffset (double off, AcDbObjectId ltId, Adesk::UInt16 color, AcDb::LineWeight lwt) ;
	virtual Acad::ErrorStatus deleteOffset (Adesk::UInt32 index) ;
	virtual Adesk::UInt32 numOffsets () ;
	virtual Acad::ErrorStatus setLinetypeOf (Adesk::UInt32 index, AcDbObjectId eId) ;
	virtual Acad::ErrorStatus linetypeOf (Adesk::UInt32 index, AcDbObjectId &eId) ;
	virtual Acad::ErrorStatus setColorOf (Adesk::UInt32 index, Adesk::UInt16 color) ;
	virtual Acad::ErrorStatus colorOf (Adesk::UInt32 index, Adesk::UInt16 &color) ;
	virtual Acad::ErrorStatus setLineWeightOf (Adesk::UInt32 index, AcDb::LineWeight lw) ;
	virtual Acad::ErrorStatus lineWeightOf (Adesk::UInt32 index, AcDb::LineWeight &lw) ;

} ;

#ifdef OPLINE_MODULE
ACDB_REGISTER_OBJECT_ENTRY_AUTO(asdkOPLineStyle)
#endif
