// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// Cyrille Fauvel
// Autodesk Developer Network

//-------------------------------------------------------------------------
//- asdkRxEventReactor.cpp : Implementation of asdkRxEventReactor
//-------------------------------------------------------------------------
#include "StdAfx.h"
#include "asdkRxEventReactor.h"

#include "asdkOPLineReactor.h"

//-------------------------------------------------------------------------
asdkRxEventReactor::asdkRxEventReactor (const bool autoInitAndRelease) : AcRxEventReactor(), mbAutoInitAndRelease(autoInitAndRelease) {
	if ( autoInitAndRelease ) {
		if ( acrxEvent )
			acrxEvent->addReactor (this) ;
		else
			mbAutoInitAndRelease =false ;
	}
}

//-------------------------------------------------------------------------
asdkRxEventReactor::~asdkRxEventReactor () {
	Detach () ;
}

//-------------------------------------------------------------------------
void asdkRxEventReactor::Attach () {
	Detach () ;
	if ( !mbAutoInitAndRelease ) {
		if ( acrxEvent ) {
			acrxEvent->addReactor (this) ;
			mbAutoInitAndRelease =true ;
		}
	}
}

void asdkRxEventReactor::Detach () {
	if ( mbAutoInitAndRelease ) {
		if ( acrxEvent ) {
			acrxEvent->removeReactor (this) ;
			mbAutoInitAndRelease =false ;
		}
	}
}

AcRxEvent *asdkRxEventReactor::Subject () const {
	return (acrxEvent) ;
}

bool asdkRxEventReactor::IsAttached () const {
	return (mbAutoInitAndRelease) ;
}

//-------------------------------------------------------------------------
void asdkRxEventReactor::beginDeepCloneXlation (AcDbIdMapping &objMap, Acad::ErrorStatus *es) {
	if ( objMap.deepCloneContext () != AcDb::kDcWblkObjects ) {
		//- If a sub-pline is copied, make sure to copy all
		//- Do it first for optimization...
		AcDbIdMappingIter iter2 (objMap) ;
		AcDbIdPair idPair2 ;
		for ( iter2.start () ; !iter2.done () ; iter2.next () ) {
			if ( !iter2.getMap (idPair2) )
				continue ;
			if ( !idPair2.isCloned () )
				continue ;

			AcDbObjectPointer<AcDbPolyline> pLine (idPair2.key (), AcDb::kForRead) ;
			if ( pLine.openStatus () != Acad::eOk )
				continue ;
			AcDbObjectPointer<AcDbPolyline> pLineClone (idPair2.value (), AcDb::kForRead) ;

			AcDbVoidPtrArray *ptr =pLine->reactors () ;
			for ( int i =0 ; i < ptr->length () ; i++ ) {
				if ( acdbIsPersistentReactor (ptr->at (i)) ) {
					//- May be one of our reactor is in?
					AcDbObjectId idr =acdbPersistentReactorObjectId (ptr->at (i)) ;
					AcDbObjectPointer<asdkOPLineReactor> pReac (idr, AcDb::kForRead) ;
					if ( pReac.openStatus () != Acad::eOk )
						continue ;

					//- Main entity
					AcDbDictionaryPointer pExtDct (pReac->ownerId (), AcDb::kForRead) ;
					if ( pExtDct.openStatus () != Acad::eOk )
						continue ;
					AcDbObjectId mainEntId =pExtDct->ownerId () ;

					AcDbIdPair pair (mainEntId, AcDbObjectId::kNull, true) ;
					if ( !objMap.compute (pair) ) {
						AcDbObjectPointer<AcDbPolyline> pMainEnt (mainEntId, AcDb::kForRead) ;
						AcDbObjectPointer<AcDbObject> pOwner (pLineClone->ownerId (), AcDb::kForWrite) ;
						if ( pOwner.openStatus () == Acad::eOk ) {
							AcDbPolyline *pMainClone =NULL ;
							if ( objMap.deepCloneContext () == AcDb::kDcWblock )
								pMainEnt->wblockClone (pOwner, (AcDbObject *&)pMainClone, objMap, true) ;
							else
								pMainEnt->deepClone (pOwner, (AcDbObject *&)pMainClone, objMap, true) ;
							pMainClone->close () ;
						}
					}
					break ;
				}
			}
			//- Do not do the following...
			//delete [] ptr ;
		}

		//- We need to copy all sub-pline(s)
		AcDbIdMappingIter iter (objMap) ;
		AcDbIdPair idPair ;
		for ( iter.start () ; !iter.done () ; iter.next () ) {
			if ( !iter.getMap (idPair) )
				continue ;
			if ( !idPair.isCloned () )
				continue ;
			
			AcDbObjectPointer<asdkOPLineReactor> pReac (idPair.key (), AcDb::kForRead) ;
			if ( pReac.openStatus () != Acad::eOk )
				continue ;

			//- Main entity
			AcDbDictionaryPointer pExtDct (pReac->ownerId (), AcDb::kForRead) ;
			if ( pExtDct.openStatus () != Acad::eOk )
				continue ;
			AcDbObjectId mainEntId =pExtDct->ownerId () ;

			AcDbIdPair idp (mainEntId, AcDbObjectId::kNull, true) ;
			objMap.compute (idp) ;
			AcDbObjectPointer<AcDbObject> pMainClone (idp.value (), AcDb::kForRead) ;

			AcDbObjectId id =pMainClone->ownerId () ;
			AcDbObjectPointer<AcDbObject> pOwnerClone (id, AcDb::kForWrite) ;

			//- In all cases force a clone of linked plines
			//- We can use IDs from the clone since there are not yet XLATED
			for ( int i =0 ; i < pReac->entsToUpdate ().length () ; i++ ) {
				if ( !objMap.compute (AcDbIdPair (pReac->entsToUpdate ().at (i), AcDbObjectId::kNull, true)) ) {
					//- Open original object
					AcDbObjectPointer<AcDbPolyline> pLine (pReac->entsToUpdate ().at (i), AcDb::kForRead) ;

					AcDbPolyline *pLineClone =NULL ;
					if ( objMap.deepCloneContext () == AcDb::kDcWblock )
						pLine->wblockClone (pOwnerClone, (AcDbObject *&)pLineClone, objMap, true) ;
					else
						pLine->deepClone (pOwnerClone, (AcDbObject *&)pLineClone, objMap, true) ;
					pLineClone->close () ;
				}
			}
		}
	}
}
