// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//
// Cyrille Fauvel
// Autodesk Developer Network

//-----------------------------------------------------------------------------
//----- asdkEdReactor.cpp : Implementation of asdkEdReactor
//-----------------------------------------------------------------------------
#include "StdAfx.h"
#include "asdkEdReactor.h"

#include "../OPLine/asdkOPLineReactor.h"

//-----------------------------------------------------------------------------
asdkEdReactor::asdkEdReactor (const bool autoInitAndRelease) : AcEditorReactor(), mbAutoInitAndRelease(autoInitAndRelease) {
	if ( autoInitAndRelease ) {
		if ( acedEditor )
			acedEditor->addReactor (this) ;
		else
			mbAutoInitAndRelease =false ;
	}
}

//-----------------------------------------------------------------------------
asdkEdReactor::~asdkEdReactor () {
	Detach () ;
}

//-----------------------------------------------------------------------------
void asdkEdReactor::Attach () {
	Detach () ;
	if ( !mbAutoInitAndRelease ) {
		if ( acedEditor ) {
			acedEditor->addReactor (this) ;
			mbAutoInitAndRelease =true ;
		}
	}
}

void asdkEdReactor::Detach () {
	if ( mbAutoInitAndRelease ) {
		if ( acedEditor ) {
			acedEditor->removeReactor (this) ;
			mbAutoInitAndRelease =false ;
		}
	}
}

AcEditor *asdkEdReactor::Subject () const {
	return (acedEditor) ;
}

bool asdkEdReactor::IsAttached () const {
	return (mbAutoInitAndRelease) ;
}

//-----------------------------------------------------------------------------
void asdkEdReactor::commandWillStart (const ACHAR *cmdStr) {
	if (   _tcsicmp (cmdStr, _T("REFEDIT")) == 0
		|| _tcsicmp (cmdStr, _T("REFCLOSE")) == 0
		|| _tcsicmp (cmdStr, _T("REFSAVE")) == 0
	) {
		asdkOPLineReactor::m_bActive =false ;
	}
	asdkOPLineReactor::m_pLinesModified.setLogicalLength (0) ;
}

void asdkEdReactor::commandCancelled (const ACHAR *cmdStr) {
	asdkOPLineReactor::m_bActive =true ;
	asdkOPLineReactor::m_pLinesModified.setLogicalLength (0) ;
}

void asdkEdReactor::commandEnded (const ACHAR *cmdStr) {
	asdkOPLineReactor::m_bActive =false ;
	for ( int i =0 ; i < asdkOPLineReactor::m_pLinesModified.length () ; i +=2 ) {
		AcDbObjectId idReac =asdkOPLineReactor::m_pLinesModified.at (i) ;
		AcDbObjectId idObj =asdkOPLineReactor::m_pLinesModified.at (i + 1) ;

		if ( !idObj.isValid () || idObj.isNull () )
			continue ;
		try {
			AcDbObjectPointer<AcDbObject> pObj (idObj, AcDb::kForRead/*Notify*/, Adesk::kTrue) ;
			AcDbObjectPointer<asdkOPLineReactor> pReac (idReac, AcDb::kForRead, Adesk::kTrue) ;
			if ( pReac.openStatus () == Acad::eOk )
				pReac->modify (pObj) ;
		} catch (...) {
		}
	}
	asdkOPLineReactor::m_pLinesModified.setLogicalLength (0) ;
	asdkOPLineReactor::m_bActive =true ;
}
